/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Originally written by Alphex Kaanoken <kaanoken@crew.org.ru>
 *
 *  NOTE: header for convertations between languages and 
 *  upper and lower cases, I'm detect only lowercases written words,
 *  and for information I'm working with char*, instead wchar_t*
 *
 *
 *  Originally written by Alfeiks Kaänoken <akaanoken@gmail.com>
 *
 *  Crew IT Research labs 2003-2006 (c)
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#include "utils.h"
#include "datastructs.h"

/* when you adding data, remember that you just give the 
 * pointer to your data and you must worry about malloc for 
 * this.
 * Actually the remove function freeing the memory area, 
 * ofcouse remove function check if this null, but
 * be careful with that.
 */
_llelem *AddItem(_llelem *p,void *data,int lpp)
{
  _llelem *lp=p;
  int id=0;
  if (p!=NULL) {
    ++id;
    while (p->next != NULL) {
      p=p->next;
      ++id;
    }
    p->next=(_llelem *) xnmalloc(sizeof(struct __llelem));
    p=p->next;
    p->next=NULL;
    p->id=id;
    p->data=data;
    return lp;
  }
  else {
    p=(_llelem *) xnmalloc(sizeof(struct __llelem));
    p->next=NULL;
    p->id=id;
    p->data=data;
    return p;
  }
}

_llelem *DelItem(_llelem *p,void *data,int id)
{
  _llelem *prev=NULL;
  _llelem *curr;
  curr=p;
  
  if (id!=-1) {
    while(curr != NULL && curr->id != id) {
      prev=curr;
      curr=curr->next;
    }
    if (curr==NULL)
      return p;
    if (curr==p)
      p=curr->next;
    else
      prev->next=curr->next;
    if (curr->data!=NULL)
      free (curr->data);
    free (curr);
    return p;
  } else if (data!=NULL) {
    while (curr!=NULL && curr->data!=data) {
      prev=curr;
      curr=curr->next;
    }
    if (curr==NULL)
      return p;
    if (curr==p)
      p=curr->next;
    else
      prev->next=curr->next;
    if (curr->data!=NULL)
      free(curr->data);
    free(curr);
    return p;
  } else {
    return p;
  }

  return p;
}

/*returns 0 if compare and -1 if not and -2 if error*/
int ElemCompare(llist *p, void *key1,void *key2)
{
  _llelem *list=p->l;

  if(list==NULL)
    return -2;

  /*we are use C String*/
  if (p->type=CSTR_TYPE) {
    while (list!=NULL && strcmp((const char*)key1,
				(const char*)list->data)!=0) {
      list=list->next;
    }
    if (list==NULL)
      return -1;
    return 0;
  }

  if (p->type=DATA_TYPE) {
    while (list!=NULL || key1!=list->data || key2!=list->data) {
      list=list->next;
    }
    if(list==NULL)
      return -1;
    return 0;
  }

  return -1;
}

llist *DestroyLList(llist *p)
{
  _llelem *list=p->l;

  if(p==NULL)
    return NULL;
  while (list!=NULL) {
    list=p->del(list,list->data,-1);
    list=list->next;
  }
  free(p);

  return NULL;
}

llist *llist_init(int type) {
  llist *p=NULL;
  p=(llist *)xnmalloc(sizeof(struct _llist));

  p->type=type;
  p->l=NULL;
  p->add=AddItem;
  p->del=DelItem;
  p->compare=ElemCompare;
  p->destroy=DestroyLList;

  return p;
}

/*binary tree operations*/
void TreeInit(Tree *tree, void (*destroy)(void *data)) 
{
  tree->size = 0;
  tree->destroy = destroy;
  tree->root = NULL;
  
  return;
}

void TreeDestroy(Tree *tree) 
{
  TreeRemLeft(tree, NULL);
  memset(tree, 0, sizeof(Tree));

  return;
}

int TreeInsLeft(Tree *tree,TreeNode *node,const void *data) 
{
  TreeNode *new_node,**position;

  if (node == NULL) {
    if (TREESIZE(tree) > 0)
      return -1;
    position = &tree->root;
  }else {
    if (TREELEFT(node) != NULL)
     return -1;

    position = &node->left;
  }

  new_node=(TreeNode *) xnmalloc(sizeof(TreeNode));
  new_node->data=(void *)data;
  new_node->left=NULL;
  new_node->right=NULL;
  *position=new_node;
  
  tree->size++;

  return 0;
}

int TreeInsRight(Tree *tree,TreeNode *node,const void *data) 
{
  TreeNode *new_node,**position;

  if(node==NULL) {
    if(TREESIZE(tree)>0)
      return -1;
    position = &tree->root;
  }else {
    if (TREERIGHT(node) != NULL)
      return -1;
    position = &node->right;
  }

  new_node=(TreeNode *) xnmalloc(sizeof(TreeNode));
  new_node->data=(void *)data;
  new_node->left=NULL;
  new_node->right=NULL;
  *position=new_node;

  tree->size++;

  return 0;
}

void TreeRemLeft(Tree *tree,TreeNode *node) 
{
  TreeNode **position;

  if (TREESIZE(tree) == 0)
    return;
  
  if (node==NULL)
    position = &tree->root;
  else
    position = &node->left;
  
  if(*position != NULL) {
    TreeRemLeft(tree,*position);
    TreeRemRight(tree,*position);
    if (tree->destroy != NULL)
      tree->destroy((*position)->data);
    
    free(*position);
    *position = NULL;
    
    tree->size--;
  }
  
  return;
}

void TreeRemRight(Tree *tree, TreeNode *node) 
{
  TreeNode **position;

  if (TREESIZE(tree) == 0)
    return;

  if(node==NULL)
    position=&tree->root;
  else
    position=&node->right;

  if(*position!=NULL) {
    TreeRemLeft(tree,*position);
    TreeRemRight(tree,*position);
    
    if(tree->destroy!=NULL) 
      tree->destroy((*position)->data);
    
    free(*position);
    *position = NULL;
    
    tree->size--;
  }
  
  return;
}

int TreeMerge(Tree *merge,Tree *left,Tree *right,const void *data) 
{
  TreeInit(merge, left->destroy);

  if(TreeInsLeft(merge,NULL,data)!=0){
    TreeDestroy(merge);
    return -1;
  }

  TREEROOT(merge)->left=TREEROOT(left);
  TREEROOT(merge)->right=TREEROOT(right);

  merge->size=merge->size+TREESIZE(left)+TREESIZE(right);

  left->root = NULL;
  left->size = 0;
  right->root = NULL;
  right->size = 0;
  
  return 0;
}

int AddTreeVoid(Tree *tree,void *data)
{
  TreeNode *prev,*node;
  int direction=0;
  void *rdata; 

  node=tree->root;
  while(TREEISEOB(node)){
    prev=node;
    if(data==(void *)TREEDATA(node))
      return -1;
    else if (data < (void *)TREEDATA(node)){
      node=TREELEFT(node);
      direction=1;
    } else {
      node=TREERIGHT(node);
      direction=2;
    }
  }

  rdata=data;

  switch(direction) {
  case 0:
    return TreeInsLeft(tree,NULL,rdata);
    break;
  case 1:
    return TreeInsLeft(tree,prev,rdata);
    break;
  case 2:
    return TreeInsRight(tree,prev,rdata);
    break;
  default:
    return -1;
    break;
  }

  return -1;
}
