/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Originally written by Alphex Kaanoken <kaanoken@crew.org.ru>
 *
 *  NOTE: header for convertations between languages and 
 *  upper and lower cases, I'm detect only lowercases written words,
 *  and for information I'm working with char*, instead wchar_t*
 *
 *
 *  Originally written by Alfeiks Kaänoken <akaanoken@gmail.com>
 *
 *  Crew IT Research labs 2003-2006 (c)
 */

#ifndef _DATASTRUCTS_H_
#define _DATASTRUCTS_H_


#include <stdio.h>
#include <stdlib.h>

/*linked list structure*/

#define CSTR_TYPE  2
#define DATA_TYPE  1

typedef struct __llelem {
  int id; /*for compatibility with old code*/
  void *data;
  struct __llelem *next;
} _llelem;

typedef struct _llist {
  int type; 
  /*I was made this for some features, because we can use list only for C strings*/
  _llelem *l;
  _llelem* (*add)(_llelem *p,void *data,int id);
  _llelem* (*del)(_llelem *p,void *data,int id);
  int (*compare)(struct _llist *p,void *key1,void *key2);
  struct _llist* (*destroy)(struct _llist *p);
} llist;

llist *init_llist(int type);

/*binary trees structures*/

/*macros */
#define TREESIZE(tree) ((tree)->size)
#define TREEROOT(tree) ((tree)->root)
#define TREEISEOB(node) ((node) == NULL)
#define TREEISLEAF(node) ((node)->left == NULL && (node)->right == NULL)
#define TREEDATA(node) ((node)->data)
#define TREELEFT(node) ((node)->left)
#define TREERIGHT(node) ((node)->right)

typedef struct _TreeNode {
  void *data;
  struct _TreeNode *left;
  struct _TreeNode *right;
} TreeNode;

#if 0
typedef struct _tree {
  int size;
  int (*compare)(const void *key1, const void *key2);
  void (*destroy)(void *data);
  /*-base tree functions*/
  void* (*init)(struct _tree *tree, void* (*destroy)(void *data));
  int (*insleft)(struct _tree *tree,TreeNode *node,const void *data);
  int (*insright)(struct _tree *tree,TreeNode *node,const void *data);
  void (*rmleft)(struct _tree *tree,TreeNode *node);
  void (*rmright)(struct _tree *tree, TreeNode *node);
  int (*merge)(struct _tree *merge,TreeNode *left,TreeNode *right,const void *data);
  /*--misc tree functions*/
  int (*addvoid)(struct _tree *tree,void *data);

  TreeNode *root;
} Tree;
#endif
typedef struct _Tree {
  int size;
  int (*compare)(const void *key1, const void *key2);
  void (*destroy)(void *data);
  TreeNode *root;
} Tree;


Tree* init_tree(Tree *p);

/*Public functions:*/
/*-base tree functions*/
void TreeInit(Tree *tree, void (*destroy)(void *data));
int TreeInsLeft(Tree *tree,TreeNode *node,const void *data);
int TreeInsRight(Tree *tree,TreeNode *node,const void *data);
void TreeRemLeft(Tree *tree,TreeNode *node);
void TreeRemRight(Tree *tree, TreeNode *node);
int TreeMerge(Tree *merge,Tree *left,Tree *right,const void *data);
/*--misc tree functions*/
int AddTreeVoid(Tree *tree,void *data);

#endif

