/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Originally written by Alphex Kaanoken <kaanoken@crew.org.ru>
 *
 *  NOTE: is a parser functions , initialize and parse
 *
 *  (c) Crew IT Research Labs 2004
 *
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "utils.h"
#include "xconfig.h"
#include "error.h"


/*linked list operations*/
app_list *AddItem(app_list * ptr, char *name,int lpp)
{
  app_list *lp=ptr;
  int id=0;
  if (ptr!=NULL) {
    ++id;
    while (ptr->next != NULL) {
      ptr=ptr->next;
      ++id;
    }
    ptr->next=(app_list *) xnmalloc(sizeof(app_list));
    ptr=ptr->next;
    ptr->next=NULL;
    ptr->id=id;
    ptr->name=strdup((const char *) name);
    ptr->lpp=lpp;
    return lp;
  }
  else {
    ptr=(app_list *) xnmalloc(sizeof(app_list));
    ptr->next=NULL;
    ptr->id=id;
    ptr->name=strdup ((const char *) name);
    ptr->lpp=lpp;
    return ptr;
  }
}

app_list *DelItem(app_list * ptr, int id)
{
  app_list *prev=NULL;
  app_list *curr;
  curr=ptr;
  
  while(curr != NULL && curr->id != id) {
    prev=curr;
    curr=curr->next;
  }
  if (curr==NULL)
    return ptr;
  if (curr==ptr)
    ptr=curr->next;
  else
    prev->next=curr->next;
  free (curr->name);
  free (curr);
  return ptr;
}

app_list *FreeItems(app_list * ptr)
{
  if (ptr!=NULL) {
    while(ptr != NULL)
      ptr=DelItem (ptr, ptr->id);
  }
  return ptr;
}

/*copy file,read file and parse it functions*/
int xconf_copy_file(const char *src_path,const char *dst_path);
int xconf_copy_file(const char *src_path,const char *dst_path)
{
  char* buf;
  int buf_size;
  int dest_desc;
  int source_desc;
  struct stat sb;
  struct stat sb_t;
  char *cp;
  int *ip;
  int return_val=0;
  off_t n_read_total=0;
  int last_write_made_hole=0;
  int make_holes=0;
  mode_t mask;

  source_desc=open(src_path,O_RDONLY);
  if (source_desc<0){
#ifdef DEBUG
    fprintf(stderr,"[ERR]: Cannot open source file: `%s'\n",src_path);
#endif
    return -1;
  }
  if(fstat(source_desc,&sb_t)){
#ifdef DEBUG
    fprintf(stderr,"[ERR]: Cannot stat source file: `%s'\n",src_path);
#endif
    return -1;
  }
  mask=sb_t.st_mode;

  dest_desc=open(dst_path,O_WRONLY | O_TRUNC | O_CREAT,mask);
  if(dest_desc<0){
#ifdef DEBUG
    fprintf(stderr,"[ERR]: Cannot create new file: `%s'\n",dst_path);
#endif
    return_val=-1;
    goto close_src_desc;
  }

  if (fstat (dest_desc, &sb)){
#ifdef DEBUG
    fprintf(stderr,"[ERR]: Cannot read status of: `%s'\n",dst_path);
#endif
    return_val = -1;
    goto close_src_and_dst_desc;
  }

  buf_size=2048;
  buf=malloc(buf_size+sizeof(int));
  if(!buf){
#ifdef DEBUG
    fprintf(stderr,"[ERR]: Cannot allocate %d bytes for buffer.\n",
	    buf_size+sizeof(int));
#endif
    return_val=-1;
    goto close_src_and_dst_desc;
  }

  for (;;){
    ssize_t n_read=read (source_desc, buf, buf_size);
    if (n_read < 0){
#ifdef DEBUG
      fprintf(stderr,"[ERR]: Error while reading source: `%s'\n",src_path);
#endif
      return_val=-1;
      goto close_src_and_dst_desc;
    }
    if (n_read == 0)
      break;/*the end (EOF)*/
    n_read_total += n_read;
    ip=0;
    if (make_holes){
      buf[n_read]=1;
      ip=(int *)buf;
      while(*ip++ ==0)/*reduce*/
	;
      /*Find first non-zero symbol*/
      cp=(char *) (ip - 1);
      while(*cp++ ==0)/*reduce*/
	;
      /* if we found its it rules,we can make a hole*/
      if(cp>buf + n_read){
	/*make hole-sucks with lseek*/
	if(lseek(dest_desc,(off_t)n_read,SEEK_CUR)<0L){
#ifdef DEBUG
	  fprintf(stderr,"[ERR]: Cannot seek in ...\n");
#endif
	  return_val=-1;
	  goto close_src_and_dst_desc;
	}
	last_write_made_hole=1;
      }
      else
	ip=0;
    }
    if(ip==0){
      size_t n = n_read;
      if(write(dest_desc,buf,n)!=n){
#ifdef DEBUG
	fprintf(stderr,"[ERR]: Cannot write to file: `%s'\n",dst_path);
#endif
	return_val=-1;
	goto close_src_and_dst_desc;
      }
      last_write_made_hole = 0;
    }
  }
  /*FIXME: ??? if we make a hole in last write, we need put to file
   * something at end*/
 close_src_and_dst_desc:
  free(buf);
  if (close(dest_desc)<0){
#ifdef DEBUG
    fprintf(stderr,"[ERR]: Cannot close destination!\n");
#endif
    return_val=-1;
  }
 close_src_desc:
  if (close (source_desc) < 0){
#ifdef DEBUG
    fprintf(stderr,"[ERR]: Cannot close the source!\n");
#endif
    return_val = -1;
  }

  return return_val;
}

xconf *ParseLine(xconf * ptr,char *line);
xconf *ParseLine(xconf * ptr,char *line)
{
  if(!line)
    return ptr;

  int len=strlen((const char*)line)+1;
  char *f_word = (char *) xnmalloc (len);
  char *word = (char *) xnmalloc (len);

  int cp = 0;
  int j = -1;
  int i = 0;
  status action, key, modifier;

  if(len<=1){
    free(word);
    return ptr;
  }

  while (line[i] != ' ') {
		word[i] = line[i];
		++i;
	}
  word[i] = NULLSYM;	

  if(!strcmp(word,"DefaultMode"))
    cp=1;
  if(!strcmp(word,"ConvPoint"))
    cp=2;
  if(!strcmp(word,"AddApp"))
    cp=3;
  if(!strcmp(word,"AddBind"))
    cp=4;
  if(!strcmp(word,"ScanDelay"))
    cp=5;
  if(!strcmp(word,"VerboseLevel"))
    cp=6; 
  if(!strncmp(word,"#",1))
    cp=7;
  if(!strcmp (word, "LangDetectMode"))
	cp=8;
  if(!strcmp (word, "EngXkbGroup"))
	cp=9;
  if(!strcmp (word, "RusXkbGroup"))
	cp=10;  
  if (!cp) {
    fprintf (stderr, "[WRN]: Unrecognized option '%s' in config file!\nIgnoring ...\n", word);
    free(word);
    return ptr;
  }
  
  /*going and getting to next word*/
  word = memset (word, NULLSYM, strlen (line) + 1);
  while (line[++i] != NULLSYM)
    word[++j] = line[i];
  word[++j] = NULLSYM;
  
  switch (cp) {
  case 1: // Get Default Mode (Manual/Auto)
    if ((ptr->DefaultMode = atoi (word)) > 1 || 
	(ptr->DefaultMode = atoi (word)) < 0) {
      fprintf (stderr,
	       "[WRN]: Bad value '%d' (use 0 or 1 for default mode), using default.\n",
	       atoi (word));
      ptr->DefaultMode = 0;
    }
    break;
  case 2: // Get Convertation Point
    if ((ptr->ConvPoint = atoi (word)) < 30 ||
	(ptr->ConvPoint = atoi (word)) > 300) {
      fprintf (stderr,
	       "[WRN]: Bad value '%d' (values must be in 30-300), using default.\n",
	       atoi (word));
      ptr->ConvPoint = 55;
    }
    break;
  case 3: // Get Applications Names
    if(!strlen(word)){
      fprintf(stderr,"[WRN]: None application WM_CLASS defined.\nIgnoring ...\n");
      free(word);
      return ptr;
    }
	int lpp = strlen(word);
	if(lpp >= 0 || lpp <= 100) {
      ptr->alist = AddItem (ptr->alist,word,lpp);
	}  
    else {
      fprintf(stderr,"[WRN]: Illegal value for application name or not defined.\nDon't use value.\n");		
    }
    break;
  case 4: // Get Keyboard Binds
	// Get Action  
	i=-1;  
    j=-1;
    f_word = memset (f_word, NULLSYM, strlen (line)+1);
	while (word[++i] != ' ')
      f_word[++j] = word[i]; 
    f_word[++j] = NULLSYM;	
	if(!strcmp(f_word,"None") || !strcmp(f_word,"0"))
      action=0;
    else if(!strcmp(f_word,"Undo") || !strcmp(f_word,"1"))
      action=1;
    else if(!strcmp(f_word,"ExLastStr") || !strcmp(f_word,"2"))
      action=2;
    else if(!strcmp(f_word,"ChgMode") || !strcmp(f_word,"3"))
      action=3;
	else if(!strcmp(f_word,"ChgSelected") || !strcmp(f_word,"4"))
      action=4;
    else{
      fprintf(stderr,"[WRN]: Invalid action name `%s'.\n",f_word);
      free(word);
      return ptr;
    }

    // Get key
    j=-1;
    f_word = memset (f_word, NULLSYM, strlen (line)+1);
    ++i;	
	while (word[i] != ' ') {
      f_word[++j] = word[i]; 
	  ++i;	
	}	
    f_word[++j] = NULLSYM;
	key = -1;
	if(!strcmp(f_word,"None") || !strcmp(f_word,"0"))
      key=0;
    if(!strcmp(f_word,"Break") || !strcmp(f_word,"Pause") || !strcmp(f_word,"1"))
      key=1;
    if(!strcmp(f_word,"ScrollLock") || !strcmp(f_word,"2"))
      key=2;
    if(!strcmp(f_word,"PrtSc") || !strcmp(f_word,"SysRq") || !strcmp(f_word,"3"))
      key=3;
    if (key < 0){
      fprintf(stderr,"[WRN]: Invalid key name `%s'.\n",f_word);
      free(word);
      return ptr;
    }

	// Get Modifier
    j=-1;
    f_word = memset (f_word, NULLSYM, strlen (line)+1);	
	++i;	
	while ((word[i] != NULLSYM)) {
      f_word[++j] = word[i]; 
	  ++i;	
	}
    f_word[++j] = NULLSYM;
	modifier = -1;
	if(!strcmp(f_word,"None") || !strcmp(f_word,"0"))
      modifier=0;
    if(!strcmp(f_word,"Shift") || !strcmp(f_word,"10"))
      modifier=10;
    if(!strcmp(f_word,"Ctrl") || !strcmp(f_word,"11"))
      modifier=11;
    if (modifier < 0){
      fprintf(stderr,"[WRN]: Invalid modifier name `%s'.\n",f_word);
      free(word);
      return ptr;
    }
	
	ptr->btable[action].key = key;
    ptr->btable[action].key_modifier = modifier;
    break;

  case 5: // Get Scaning Delay
    if( ((ptr->ScaningDelay = atoi (word)) < 1000 )
	|| (ptr->ScaningDelay > 30000 ) ){
      send_error(WARNING,
		 "ScaningDelay is set to not recommended value %d \n"	\
		 "(Recommended: 1000-30000). Set to default.\n",
		 &(ptr->ScaningDelay) );
      ptr->ScaningDelay = DEFAULT_SCANNING_DELAY;
    }
    break;
  case 6: // Get Debug Mode
    if(!strcmp(word,"Silent"))
      ptr->VerboseMode=0;
    else if(!strcmp(word,"Debug"))
      ptr->VerboseMode=1;
    else if(!strcmp(word,"DevDebug"))
      ptr->VerboseMode=2;
    else {
      fprintf(stderr,"[WRN]: Illegal option for debug level `%s'.\n",word);
      free(word);
      return ptr;
    }
    break;
  case 7:
	break;	
  case 8:
	if ((ptr->LangDetectMode = atoi (word)) > 1
	    || (ptr->LangDetectMode = atoi (word)) < 0) {
	  fprintf (stderr,
	  "[WRN]: Bad value '%d' (use 0 or 1 for lang detection mode), using default.\n",
	  atoi (word));
	  ptr->LangDetectMode = 0;
	}
	break;	
  case 9:	  
	ptr->xkbGroup[0] = atoi (word);
	break;	
  case 10:
	ptr->xkbGroup[1] = atoi (word);
	break;	
  default:
    fprintf (stderr, "[WRN]: Ignored value '%s'.\n", word);
    break;
  }
  
  free (word);
  return ptr;
}

int xconf_create_file (xconf * ptr, const char *fl_name);
int xconf_create_file (xconf * ptr, const char *fl_name)
{
  FILE *stream = NULL;
  if ((stream = fopen (fl_name, "w")) != NULL) {
    fputs("# It's a default X Neural Switcher configuration file\n", stream);
    fputs("# all values are default\n\n", stream);
    fputs("# default mode in window, 0 is manual, 1 is automatically\n", stream);
    fputs("DefaultMode 0\n", stream);
    fputs("# it's a point for convertation words\n", stream);
    fputs("# recomendated value is 55, You can play with this value for\n", stream);
    fputs("# # custom rules.\n", stream);
    fputs("ConvPoint 55\n\n", stream);
    fputs("# Add Applications names for procces with xneur\n", stream);
    fputs("# Only for listed applications (it's window names)\n", stream);
    fputs("# xneur will process the input\n", stream);
    fputs("# Example:\n", stream);
    fputs("#AddApp Sylpheed\n\n", stream);
    fputs("AddApp Gedit\n", stream);
    fputs("AddApp Anjuta\n", stream);
    fputs("AddApp Sailo\n", stream);
    fputs("AddApp Gaim\n\n", stream);
	fputs("# define xkb group\n", stream);
	fputs("EngXkbGroup 0\n", stream);
	fputs("RusXkbGroup 1\n\n", stream);	  
	fputs("# Select algorithm of language detection\n", stream);
	fputs("# 0 - Algo by A. Kaanoken\n", stream);
	fputs("# 1 - Algo by A. Kuznetsov\n", stream);
	fputs("LangDetectMode 1\n", stream);
	  
    fputs("# it's all\n", stream);

    fclose (stream);
	return SUCCESS;
  }
  return ERROR;
}

int xconf_load_file (xconf * ptr, const char *fl_name);
int xconf_load_file (xconf * ptr, const char *fl_name)
{
  FILE *stream = NULL;
  int sym = 0;
  int i = -1;
  /*States of Parser */
  char *t_word = (char *) xnmalloc (258);
  
  if ((stream = fopen (fl_name, "r")) != NULL) {
    while ((sym = fgetc (stream)) != EOF) {
      switch (sym) {
        case '\n':
          if ((strlen ((const char *) t_word)) != 0) {
            t_word[++i] = NULLSYM;
            ptr=ParseLine(ptr,t_word);
            t_word=memset(t_word, NULLSYM, 258);
            i = -1;
          }
          break;
        default:
          t_word[++i] = (unsigned char) sym;
          break;
      }
      
      if (i > 255) {
	    fprintf (stderr,"[ERR]: Configuration file '%s' is bad!\nExit(1)\n", fl_name);
	    free (t_word);
	    fclose (stream);
	    exit (1);
      }
    }
    fclose (stream);
    free (t_word);
    
    return SUCCESS;
  }
  else {
    fprintf (stderr,
	     "[WRN]: Cannot read config file: %s, using default settings!\n", fl_name);
    // Cannot read, create file
    ptr->create_file(ptr, fl_name);
    free (t_word);
    return WARNING;
  }
}

int xconf_uninit (xconf * ptr);
int xconf_uninit (xconf * ptr)
{
  ptr->alist=FreeItems(ptr->alist);
  free (ptr);
  return SUCCESS;
}

char* xconf_load_dict_file (const char *fl_name);
char* xconf_load_dict_file (const char *fl_name)
{
	FILE *stream = NULL;
	int sym = 0;
	int i = 0;
	// States of Parser 
	int isRead = 1;
	int t = 0;
	char *t_word = (char *) xnmalloc (258);
	char *word = (char *) xnmalloc (1);
	// init russian dictionary
	if ((stream = fopen (fl_name, "r")) != NULL) {
		word[0]='\0';
		t_word[i] = '|';
		t_word[++i] = '\n';
		while ((sym = fgetc (stream)) != EOF) {
			switch (sym) {
				case '#':
					isRead = 0;
					break;
				case '\n':
					if ((t = strlen ((const char *) t_word)) != 0)
						isRead = 10;
					else
						isRead = 0;
					if (isRead) {
						t_word[i] = '|';
						t_word[++i] = NULLSYM;
						word = realloc(word, strlen((const char *)t_word) + strlen((const char *)word) + 1);
						strcat(word, t_word);
						//printf ("'%s'\n", word);
						t_word = memset (t_word, NULLSYM, 258);
						i = 0;
					}
					isRead = 1;
					break;
				default:
					if (isRead) {
						t_word[i] = (unsigned char) sym;
						t_word[++i] = NULLSYM;
						//printf ("'%s'\n", t_word);
					}
					break;
			}

			if (i > 255) {
				fprintf (stderr,
					 "[ERR]: Russian Dictionary file '%s' is bad!\nExit(1)\n");
				free (t_word);
				fclose (stream);
				exit (1);
			}
		}
		fclose (stream);	
		free (t_word);		
		return word;
	}
	else {
		fprintf (stderr,
			 "[WRN]: Cannot read russian dictionary file: %s, russian dictionary not use!\n", fl_name);
		if (t_word != NULL) 
			free (t_word);
		return NULL;
	}
}

xconf *xconf_init (xconf * ptr)
{
  ptr = (xconf *) xnmalloc (sizeof (struct _xconf));
  ptr->DefaultMode=AUTO_BY_KEYLOG;
  ptr->CurrentMode=AUTO_BY_KEYLOG;
  ptr->ConvPoint=55;	/*default values */
  ptr->LangDetectMode = 1;  /*new algorithm*/
  ptr->ScaningDelay=DEFAULT_SCANNING_DELAY;
  ptr->alist=NULL;
  ptr->aplist=NULL;
  ptr->VerboseMode=1; /*default normal debug*/
  ptr->xkbGroup[0]=0; /* default english keyboard group */
  ptr->xkbGroup[1]=1; /* default russian keyboard group */  
  
  memcpy(ptr->btable, DEFAULT_BT, DEFAULT_BIND_TABLE_SIZE*sizeof(bind));
  
  // Function mapping
  ptr->copy_file = xconf_copy_file;
  ptr->create_file = xconf_create_file;
  ptr->load_file = xconf_load_file;
  ptr->uninit = xconf_uninit;
  ptr->load_dict_file = xconf_load_dict_file;
  return ptr;
}
