/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Originally written by Alphex Kaanoken <kaanoken@crew.org.ru>
 *
 *  NOTE: is a parser header, functions prototypes and structures for X Neural Switcher 
 *
 *  (c) Crew IT Research Labs 2004
 *
 */

#ifndef _xconfig_h
#define _xconfig_h

#include "types.h"

/* Init config values*/
#define DEFAULT_SCANNING_DELAY 20000	/*microsec*/

typedef struct _app_list {
  int id;
  char *name;
  int lpp;
  struct _app_list *next;
} app_list;

typedef struct _llist {
  app_list *list;
  int id;
  app_list* (*add)(app_list *p);
  app_list* (*rm)(app_list *p);
  app_list* (*free)(app_list *p);
  app_list* (*destroy)(app_list *p);
} llist;

typedef struct _bind{
  status key;
  status key_modifier;
} bind;

static bind DEFAULT_BT[] = { {NONE_KEY,NONE_KEY},{PAUSE_KEY, NONE_KEY},{PAUSE_KEY, SHIFT_KEY},{SYS_RQ_KEY, CONTROL_KEY} };

typedef struct _xconf {
  int ConvPoint;
  status DefaultMode;
  status CurrentMode; /*AUTO_BY_KEYLOG,MANUAL_ON_SHORTCUT */
  size ScaningDelay; /*microsecs*/
  status LangDetectMode;  // 0 - old mode; 1 - new mode
  app_list *alist;
  llist *aplist;
  int VerboseMode; /*0-Silent,1-Normal debug,2-Heavy debug*/
  bind btable[DEFAULT_BIND_TABLE_SIZE];
  char *rudict;
  char *endict;
  int xkbGroup[2];	
  int (*copy_file)(const char *src_path,const char *dst_path);
  int (*create_file) (struct _xconf * ptr,const char *fl_name);
  int (*load_file) (struct _xconf * ptr,const char *fl_name);
  char* (*load_dict_file) (const char *fl_name);
  int (*uninit) (struct _xconf * ptr);
} xconf;

xconf *xconf_init (xconf * ptr);

app_list *AddItem(app_list * ptr,char *name,int lpp);
app_list *DelItem(app_list * ptr,int id);
app_list *FreeItems(app_list * ptr);

#endif /*_xconfig_h*/
