/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Originally written by Art Vodolazsky <burivuh@crew.org.ru>
 *  a lot of midifications by Alfeiks Kaänoken <akaanoken@gmail.com>
 *
 *  (c) Crew IT Research Labs 2004, 2005
 */

#ifndef _LIBRARY_H_
#define _LIBRARY_H_

#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/XKBlib.h>
#include <X11/Xutil.h>

#include "../libmisc/types.h"

int handleXErrors (Display * d, XErrorEvent * e);

/* IMPORTANT NOTES:*/
/* 1)Work only with null-terminated strings.
 *   If not - correct result is not guaranteed.
 */

/*working with xObject*/

#define CLIENT_EVENT_MASK KeyPressMask|StructureNotifyMask

#define RUSSIAN 0
#define ENGLISH 1
#define RUSWINXkbGROUP "Russia"
#define USAENGXkbGROUP "U.S. English"

#define CYR_BYTES_PER_SYM 2

typedef struct _xobj
{
  Display *xDisplay;
  num xScreenNum;
  Window xWindow;
  int (*connect)(struct _xobj *p);
  int (*create_win)(struct _xobj *p);
  int (*set_err_handler)(int (*handler) (Display *,XErrorEvent *));
  Window (*get_app_root_win)(struct _xobj *p);
  Window (*find_win_byatom)(struct _xobj *p,char *atom);
  Window (*find_child_byatom)(Display *xdsp,Window *xwin,Atom atom);
  Display* (*get_display)(struct _xobj *p);
  Window (*get_win)(struct _xobj *p);
  XClassHint* (*get_wm_class)(struct _xobj *p);
  char* (*get_wmclass_name)(struct _xobj *p);
  char* (*get_wmclass_class)(struct _xobj *p);
  int (*set_event_mask)(struct _xobj *p,mask event_mask);
  int (*uninit)(struct _xobj *p);
} xobj;

xobj *xobj_init (xobj *p);

/*working with xIFocus*/
typedef struct _coords
{
  int x;
  int y;
} coords;

typedef struct _xifocus
{ xobj *owner; //input focus window
  xobj **ptr; /*pointer windows*/
  int ptr_num;
  int cur_status;/*input focus status*/
  int changed;/*TRUE if after update owner of input focus changed. By init is FALSE*/
  coords pointerSave;
  int (*update)(struct _xifocus *p);
  bool (*has_chng)(struct _xifocus *p);
  int (*uninit)(struct _xifocus *p);
} xifocus;

xifocus *xifocus_init (xifocus *p);

/*working with xSelection*/

#define MIN_CYCLE_TIME 10

typedef struct _xselection
{
  Atom selection; /*PRIMARY,SECONDARY,CLIPBOARD*/
  Atom target; /*"recieve" format selection atom*/
  Atom property; /*"recieve" selection atom*/
  Atom encoding;
  time_t currentTime;
  xobj *owner; /*selection owner*/
  xobj *requestor; /*selection requestor*/
  int (*request_to_convert)(struct _xselection *p);
  XSelectionEvent* (*wait_for_notify)(struct _xselection *p);
  char* (*get_to_string)(struct _xselection *p,XSelectionEvent *xselt);
  unsigned char* (*get)(struct _xselection *p);
  int (*set_to_none)(struct _xselection *p);
  int (*wait_owner)(struct _xselection *p,int time);
  int (*uninit)(struct _xselection *p);
} xselection;

xselection* xselection_init(xselection *p,Display *ManagedDisplay,Window ManagedWindow);

/*working with language switching*/

typedef struct _xswitchlang
{
  xobj *xobject;
  //XkbStateRec state;
  int currlang;
  int rus_win_keyboard_group;
  int usa_eng_keyboard_group;	
  int (*switch_lang)(struct _xswitchlang *p);
  char** (*getKeyboardList) (int* length, struct _xswitchlang *p);
  void (*setActiveKeyboard) (int group, struct _xswitchlang *p);
  int (*getActiveKeyboard) (struct _xswitchlang *p);	  
  int (*get_curr_lang)(struct _xswitchlang *p, Display *dpy);
  int (*uninit)(struct _xswitchlang *p);
} xswitchlang;

xswitchlang *xswitchlang_init(struct _xswitchlang *p,Display * xDisplay);

/*working with xString*/

typedef struct _xstr
{
  char *content; /*string itself*/
  size currentSize; /*current size of content, keycode, keycodeModifiers fields*/
  KeyCode *keycode; /*array of string chars keycodes*/
  mask *keycodeModifiers; /*array of string chars keycodes modifiers*/
  xobj *home; /*current processing xobj*/
  int (*set_key_code)(struct _xstr *p,int lang, xswitchlang *swlng);
  int (*to_ascii)(struct _xstr *p); // Function not use!
  int (*set_content)(struct _xstr *p,char *src);
  int (*cat_content)(struct _xstr *p,struct _xstr *dest);
  int (*add_ch_byindx)(struct _xstr *p,char sym,int index);
  int (*set_ch_byindx)(struct _xstr *p,char sym,int index);
  char (*get_ch_byindx)(struct _xstr *p,int index);
  size_t (*get_len)(struct _xstr *p);	  
  int (*uninit)(struct _xstr *p);
} xstr;

xstr *xstr_init (xstr *p,size init_size);

/*working with hotkeys*/

typedef struct _xbinding
{
  KeySym key;
  KeySym key_alt;
  KeySym modifier;
} xbinding;

typedef struct _xbindingtable
{
  xbinding *bindlist;
  int ibind;
  int current_action;
  int (*set_action)(struct _xbindingtable *p,int action);
  int (*get_action)(struct _xbindingtable *p);
  int (*bind_key)(struct _xbindingtable *p,KeySym key,KeySym alt,
		 KeySym modifier,int action);
  int (*find_bind_action)(struct _xbindingtable *p,KeySym key,
			KeySym modifier);
  int (*uninit)(struct _xbindingtable *p);
} xbindingtable;

xbindingtable* xbindingtable_init(xbindingtable *p,int ibind);

/*working with xEvent*/

//#define CyrMask 1<<13

#define Group0Mask 0x00000000
#define Group1Mask 0x00002000 // 1<<13
#define Group2Mask 0x00004000
#define Group3Mask 0x00006000 // 1<<14
//

typedef struct _xevent
{
  XEvent event;	/*event to process*/
  xobj *communicator; /*xObject that recieves/sends event*/
  int (*set_communicator)(struct _xevent *p,xobj *set);
  Window (*get_communicator_win)(struct _xevent *p);
  Window (*get_event_win)(struct _xevent *p);
  int (*update_by_focus)(struct _xevent *p,xifocus *f);
  int (*send_xkey)(struct _xevent *p,KeyCode kc,mask modifiers);
  int (*send_event)(struct _xevent *p,XEvent *event,mask emask);
  int (*send_string)(struct _xevent *p,xstr *sp);
  int (*put_back_to_queue)(struct _xevent *p, XEvent *event);
  bool (*chck_key_release)(struct _xevent *p);
  bool (*chck_key_press)(struct _xevent *p);
  bool (*chck_destroy_notify)(struct _xevent *p);
  bool (*is_native)(struct _xevent *p);
  char (*get_curr_char)(struct _xevent *p, struct _xswitchlang *xlang);
  KeySym (*get_curr_keysym)(struct _xevent *p);
  KeySym (*get_curr_modifier)(struct _xevent *p);
  //char (*ruchar_to_engchar)(unsigned char ch);
  //KeySym (*keysym_to_latin1)(KeySym ks);
  KeySym (*char_to_keysym)(char ch);
  char (*keysym_to_char)(KeySym ks);    
  int (*uninit)(struct _xevent *p);
} xevent;

xevent* xevent_init(xevent *p);

#endif /* _XLIBRARY_H_ */
