/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Originally written by Art Vodolazsky <burivuh@crew.org.ru>
 *  a lot of modifications by Alphex Kaänoken <kaanoken@crew.org.ru>
 *  after Andrew Crew Kuznetsov <andrew.crew.kuznetsov@gmail.com> rewrite it.
 *
 *  (c) Crew IT Research Labs 2004
 */

#include "program.h"
#include "../libmisc/conv.h"
#include "../libmisc/utils.h"
#include "../libai/detection.h"

#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include <X11/Xdefs.h>
#include <X11/Xmu/Atoms.h>

/*global*/
static status exit_flag = 1;

/*set exit flag*/
void xprogram_exit()
{
  exit_flag = 0;
}

/*test exit flag*/
bool xprogram_isexiting()
{
  if (exit_flag)
    return FALSE;
  else
    return TRUE;
}

/*close connection & free xProgramInfo struct*/
int xprogram_uninit(struct _xprograminfo *p);
int xprogram_uninit(struct _xprograminfo *p)
{
  int err = SUCCESS;
  
  XDestroyWindow (p->xobject->get_display(p->xobject),
		  p->xobject->get_win(p->xobject));
  XCloseDisplay (p->xobject->get_display(p->xobject));
  
  if (p) {
    if ((p->xobject->uninit(p->xobject) != SUCCESS) ||
	(p->xstring->uninit (p->xstring) != SUCCESS) ||
	(p->xword->uninit (p->xword) != SUCCESS) ||
	(p->selection->uninit (p->selection) != SUCCESS) ||
	(p->xevnt->uninit (p->xevnt) != SUCCESS) ||
	(p->ifocus->uninit (p->ifocus) != SUCCESS) ||
	(p->xswitchlng->uninit (p->xswitchlng) != SUCCESS) ||
	(p->xklbuffer->free(p->xklbuffer) != SUCCESS) ||
	(p->xbtable->uninit (p->xbtable) != SUCCESS))
      err = WARNING;
    free (p);
  }
  else
    err = WARNING;
  
  return err;
}

/*get info about current input focus owner*/
bool xprogram_update(struct _xprograminfo *p, app_list * sup_apps_list);
bool xprogram_update(struct _xprograminfo *p, app_list * sup_apps_list)
{
  status err=SUCCESS;
  /*setting xEvent reciever params to updated focus window params*/
  err=xClientManager_getClientInfo(p->xevnt,p->ifocus);
  /*if focus owner window is set to None (ERROR) */
  if(ERROR==err)
    return FALSE;
  if(!xClientManager_testAppName(p->ifocus,sup_apps_list))
    return FALSE;
  /*begin listening application*/
  xClientManager_beginToListenClient(p->ifocus);
  return TRUE;
}

/*log keys*/
num xprogram_process_input(struct _xprograminfo *p);
num xprogram_process_input(struct _xprograminfo *p)
{
  char *res_str = NULL;
  num res_chars = 0;
  
  /*return early if bind detected*/
  res_str=xKeyLogManager_getStringFromEventQueue (p->xklbuffer,
						  p->xevnt, p->ifocus, p->xword,
						  p->xstring, p->xswitchlng,
						  p->xbtable, p->get_curr_lang(p));
  if (res_str)
    res_chars = strlen (res_str);
  else
    res_chars = 0;

  //if (res_str!=NULL) free(res_str);
	  
  return res_chars;
}

/*process bind action*/
bool xprogram_perform_current_bind(struct _xprograminfo *p, xconf* xnc);
bool xprogram_perform_current_bind(struct _xprograminfo *p, xconf* xnc)
{
  bool back_to_input = TRUE;
  status action = p->xbtable->get_action(p->xbtable);
  status lang, err;
  char sym;
  int iterator;	
  KeySym ks;
	
  if (action != BA_NONE) {
    switch (xnc->CurrentMode) {
      /*default auto mode*/
      case AUTO_BY_KEYLOG:
        switch (p->xbtable->get_action(p->xbtable)) {
          /*user need to cancel last change*/
          case BA_CHANGE_WORD:
            lang = p->get_curr_lang(p);
            if (ENGLISH == lang) {
              lang = RUSSIAN;
              send_error (DEBUG, "Back -> to Rus\n", NULL);
            }
            else {
              lang = ENGLISH;
              send_error (DEBUG, "Back -> to Eng\n", NULL);
            }
            xKeyLogManager_flushBuffer (p->xklbuffer, p->xword);  
            err = p->chng_word_by_selection (p, lang);
		    send_error (DEBUG, "Word changed by selection\n", NULL);
            back_to_input = TRUE;
			if (strlen(p->xstring->content) > 0)
              send_error (DEBUG, "Current string=%s\n", &p->xstring->content);
            break;
          case BA_CHANGE_MODE:
            xnc->CurrentMode = MANUAL_ON_SHORTCUT;
            send_error(DEBUG, "Mode -> to Manual\n", NULL);
            back_to_input = FALSE;
            break;
		  /*user need to change current phrase*/
          case BA_CHANGE_STRING:
  	        lang = p->get_curr_lang(p);
	        if (ENGLISH == lang) {
  	          lang = RUSSIAN;
	          send_error (DEBUG, "Eng -> to Rus\n", NULL);
	        }
	        else {
 	          lang = ENGLISH;
	         send_error (DEBUG, "Rus -> to Eng\n", NULL);
	        }
	        xKeyLogManager_flushBuffer (p->xklbuffer, p->xword);
	        err = p->chng_string_by_selection (p, lang);
		    p->xklbuffer->clear(p->xklbuffer);
	        back_to_input = TRUE;
			if (strlen(p->xstring->content) > 0)
              send_error (DEBUG, "Current string=%s\n", &p->xstring->content);
	        break;
		  case BA_CHANGE_SELECTED:
            send_error(DEBUG, "Change Selected Text\n", NULL);
		    unsigned char * seltext = NULL;
		    seltext = p->selection->get(p->selection);
		    if (seltext) {
			  if (strlen((char*)seltext) < MAX_STRING_LENGTH) {
		        send_error (DEBUG, "Selected Text '''%s'''\n", &seltext);
			    p->set_convert_text(p, seltext);
			  }
			  else {
		        send_error (DEBUG, "Selected Text Too Long!\n", NULL);				
			  }			  
			  XSetSelectionOwner (p->selection->requestor->xDisplay, p->selection->selection, None, CurrentTime);
			}
		    break;
          default:
            break;
        }
      break;
      /*default manual mode*/
    case MANUAL_ON_SHORTCUT:
      switch (action) {
	    /*user need to change last word*/
        case BA_CHANGE_WORD:
		  send_error (DEBUG,"Processing word: %s", &(p->xword->content));
          lang = p->get_curr_lang(p);
		  if (lang == ENGLISH)
	        send_error (DEBUG, " (language ENGLISH)\n", NULL);
	      if (lang == RUSSIAN)
	        send_error (DEBUG, " (language RUSSIAN)\n", NULL);
		  
		  if (ENGLISH == lang) {
	        lang = RUSSIAN;
	        send_error (DEBUG, "Eng -> to Rus\n", NULL);
	      }
	      else {
	        lang = ENGLISH;
	        send_error (DEBUG, "Rus -> to Eng\n", NULL);
	      }	      
	      xKeyLogManager_flushBuffer (p->xklbuffer, p->xword);
	      err = p->chng_word_by_selection (p, lang);
	      back_to_input = TRUE;
	  	  if (strlen(p->xstring->content) > 0)
            send_error (DEBUG, "Current string=%s\n", &p->xstring->content);
	      break;
	    /*user need to change current phrase*/
        case BA_CHANGE_STRING:
	      lang = p->get_curr_lang(p);
	      if (ENGLISH == lang) {
	        lang = RUSSIAN;
	        send_error (DEBUG, "Eng -> to Rus\n", NULL);
	      }
	      else {
	        lang = ENGLISH;
	        send_error (DEBUG, "Rus -> to Eng\n", NULL);
	      }
	      xKeyLogManager_flushBuffer (p->xklbuffer, p->xword);
	      err = p->chng_string_by_selection (p, lang);
		  p->xklbuffer->clear(p->xklbuffer);
	      back_to_input = TRUE;
		  if (strlen(p->xstring->content) > 0)
            send_error (DEBUG, "Current string=%s\n", &p->xstring->content);
	      break;
        case BA_CHANGE_MODE:
	      xnc->CurrentMode = AUTO_BY_KEYLOG;
	      send_error(DEBUG, "Mode -> to Auto\n", NULL);
	      back_to_input = FALSE;
	      break;
		case BA_CHANGE_SELECTED:
          send_error(DEBUG, "Change Selected Text\n", NULL);
		  unsigned char * seltext = NULL;
		  seltext = p->selection->get(p->selection);
		  if (seltext) {
			if (strlen((char*)seltext) < MAX_STRING_LENGTH) {
		      send_error (DEBUG, "Selected Text '''%s'''\n", &seltext);
			  p->set_convert_text(p, seltext);
			}
			else {
		      send_error (DEBUG, "Selected Text Too Long!\n", NULL);				
			}
			XSetSelectionOwner (p->selection->requestor->xDisplay, p->selection->selection, None, CurrentTime);
		  }
		  break;
        default:
	      break;
        }
      break;
    default:
      break;
    }
    /*notify we'll have no bind*/
    p->xbtable->set_action(p->xbtable, BA_NONE);
  }
  else
    back_to_input = FALSE;/*no bind*/
  
  return back_to_input;
}

/*perform mode actions*/
int xprogram_perform_current_mode(struct _xprograminfo *p, xconf * xnc);
int xprogram_perform_current_mode(struct _xprograminfo *p, xconf * xnc)
{
  num cnp = 0;
  char *logged_word = p->xword->content;
  //size word_len = p->xword->get_len(p->xword);
  char *low_word = NULL;
  //char *ret_word = NULL;
  status lang, err = SUCCESS;
  char sym;
  char word[66];
  int i = -1;
  int j;
	
  switch (xnc->CurrentMode) {
    /*default auto mode*/
    case AUTO_BY_KEYLOG:
      /*get current layout*/
      lang = p->get_curr_lang(p);
      send_error (DEBUG, "Processing word: %s", &logged_word);
	  if (lang == ENGLISH)
	    printf(" (language ENGLISH)\n");
	  if (lang == RUSSIAN)
	    printf(" (language RUSSIAN)\n");
	  low_word = AllToLower ((const char *) logged_word);
      if (low_word == NULL)
        return err;
	  if (xnc->LangDetectMode==0) {
        /*test word by lang algorythm*/        
        cnp = isCyrWord ((const char *) low_word);
        if ((cnp > xnc->ConvPoint) && (ENGLISH == lang)) {
          /*process string in client window*/
          err = p->chng_word_by_selection (p, RUSSIAN);
          send_error (DEBUG, "Eng -> to Rus\n", NULL);
        }
        else if ((cnp < xnc->ConvPoint) && (RUSSIAN == lang)) {
          /*process string in client window*/
          err = p->chng_word_by_selection (p, ENGLISH);
          send_error (DEBUG, "Rus -> to Eng\n", NULL);
        }
        else
          send_error (DEBUG, "Nothing to change\n", NULL);
	  }	
	  else if (xnc->LangDetectMode==1) {
		send_error (DEBUG, "Use New Algorithm (language %d)\n", &lang);
		word[++i] = '|';
		for (j = 0; j < strlen((const char*) low_word); j++){
		  if (low_word[j] != '\0')
			word[++i] = low_word[j];
		}
		word[++i] = '|';
		word[++i] = '\0';
		//printf("word: '%s'\n", word);
		//printf("dict ru: '%s'\n", xnc->rudict);
		//printf("dict en: '%s'\n", xnc->endict);
		
		// check russian dictionary 
		if ((xnc->rudict != NULL) && (xnc->endict != NULL)) {
          if (strstr(xnc->rudict, word) != NULL) 
		    cnp = 1;
		  // check english dictionary    
          if (strstr(xnc->endict, word) != NULL) 
		    cnp = -1;
	    }  
		// if not find in dictionary, use heuristic detection
        if (cnp == 0) 
		  cnp = WhatLang ((const char *) low_word);
		if ((cnp > 0) && (ENGLISH == lang)) {
		  //process string in client window
		  err = p->chng_word_by_selection (p, RUSSIAN);
          send_error (DEBUG, "Eng -> to Rus\n", NULL);;
		}
		else if ((cnp < 0) && (RUSSIAN == lang)) {
		  //process string in client window
		  err = p->chng_word_by_selection (p, ENGLISH);
          send_error (DEBUG, "Rus -> to Eng\n", NULL);
		}
		else
		  send_error (DEBUG, "Nothing to change\n", NULL);
	  }
	  else
		return err;  
	  if (strlen(p->xstring->content) > 0)
        send_error (DEBUG, "Current string=%s\n", &p->xstring->content);
      free (low_word);
      break;
    case MANUAL_ON_SHORTCUT:
      break;
    default:
      break;
  }
  return err;
}

/*change last word in window*/
int xprogram_chng_word_by_selection(struct _xprograminfo *p, status lang);
int xprogram_chng_word_by_selection(struct _xprograminfo *p, status lang)
{
  status err = SUCCESS;
  size word_len = 0;
  char *word = NULL;
  xselectmanager* xsm = NULL;

  xsm = xselectmanager_init(xsm);	

  if (!p->xword->get_len(p->xword)) {
    word = (char *)get_last_word(p->xstring->content);
    send_error (DEBUG, "Get word='%s' from string buffer\n", &word);
    if (!word)
      return WARNING;
    p->xword->set_content (p->xword, word);
    free (word);
  }
  word_len = p->xword->get_len(p->xword);
  /*delete old word*/
  err = xsm->del_nchrbck(p->xevnt, word_len);
  /*write changed word*/
  err = xsm->send_string(p->xword, p->xswitchlng, p->xevnt, lang);
  /*      xString_setContent(xinfo->xword, NULL);*/
  /*      xString_setContent(xinfo->xstring, NULL);*/
  xsm->uninit(xsm);
  return err;
}

/*change last phrase in window*/
int xprogram_chng_string_by_selection(struct _xprograminfo *p, status lang);
int xprogram_chng_string_by_selection(struct _xprograminfo *p, status lang)
{
  status err = SUCCESS;
  size str_len;
  xselectmanager* xsm = NULL;
  xsm = xselectmanager_init(xsm);

  if (p->xstring->get_len(p->xstring) == 0)
    return WARNING;
  /*cat last word buffer & current phrase buffer*/
  p->xword->cat_content(p->xword, p->xstring);
  str_len = p->xstring->get_len(p->xstring);
  
  /*delete old phrase*/
  err = xsm->del_nchrbck(p->xevnt, str_len);
  /*write changed phrase*/
  err = xsm->send_string(p->xstring, p->xswitchlng, p->xevnt, lang); 
  /*zero current buffer phrase & word*/
  //p->xstring->set_content (p->xstring, NULL);
  p->xword->set_content (p->xword, NULL);
  
  xsm->uninit(xsm);
  
  return err;
}

/*get current layout*/
int xprogram_get_curr_lang(struct _xprograminfo *p);
int xprogram_get_curr_lang(struct _xprograminfo *p)
{
  return p->xswitchlng->currlang;
}

/* set text with convert */
int xprogram_set_convert_text(xprograminfo *p, unsigned char *text);
int xprogram_set_convert_text(xprograminfo *p, unsigned char *text)
{
  xselectmanager* xsm = NULL;
  xevent* xevnt = NULL;	
  xsm = xselectmanager_init(xsm);
  xevnt = xevent_init(xevnt);
	
  mask keycode_mod;
  mask shift_mod;
	
  KeyCode *keycode; 
  mask *keycodeModifiers;				
  keycode=(KeyCode *)xnmalloc(sizeof(KeyCode) * MAX_STRING_LENGTH);
  keycodeModifiers=(mask *)xnmalloc(sizeof(mask) * MAX_STRING_LENGTH);
  memset(keycode,0,MAX_STRING_LENGTH);
  memset(keycodeModifiers,0,MAX_STRING_LENGTH);				
      				
  KeySym ks;				
  char * string = xnmalloc(strlen((const char*)text));
  strcpy(string, (const char*)text);
  int i = 0; 
  unsigned int j = -1;
  char element[2]={'\0','\0'};
  char * cyrstr = xnmalloc(3);
  char * newstr = xnmalloc(MAX_STRING_LENGTH);
  newstr = memset (newstr, NULLSYM, MAX_WORD_LENGTH);
  for(i=0; i<strlen(string); i++) {
    if (isascii(string[i])) {
      //printf("ASCII\n");		
	  newstr[++j]=string[i];
	  switch (p->xswitchlng->rus_win_keyboard_group) {
		case 0:
		  keycode_mod = Group0Mask;
	  	  break;
		case 1:
	  	  keycode_mod = Group1Mask;
	  	  break;
		case 2:
	      keycode_mod = Group2Mask;
	      break;
	    case 3:
	  	  keycode_mod = Group3Mask;
	  	  break;
  	  }
	}	
	else {
	  //printf("CYR\n");
	  cyrstr[0]=string[i];
	  i++;
	  cyrstr[1]=string[i];
	  cyrstr[2]='\0';
	  newstr[++j]=get_ascii(cyrstr);
	  if (newstr[j] == NULLSYM) {
        newstr[j]='#';
		i++;
	  }
  	  switch (p->xswitchlng->usa_eng_keyboard_group) {
		case 0:
	  	  keycode_mod = Group0Mask;
	  	  break;
		case 1:
	  	  keycode_mod = Group1Mask;
	  	  break;
		case 2:
	  	  keycode_mod = Group2Mask;
	  	  break;
		case 3:
	  	  keycode_mod = Group3Mask;
	  	  break;
  	  }		
	}	

	element[0] = newstr[j];
	if (isUpperNonAlphaCyr (element[0]) | isupper(element[0]))
  	  shift_mod = ShiftMask;
	else
	  shift_mod = None;	
		
    keycodeModifiers[j] = (keycode_mod | shift_mod);
	ks=xevnt->char_to_keysym(element[0]);
	if (ks == None)
	  ks = XK_Return;
   	keycode[j]=XKeysymToKeycode(p->xstring->home->get_display(p->xstring->home),ks);		
	// Xlib (error '<') patch >>
   	if ((ks == XK_less) && (keycode[j] == 94))
  	  keycode[j] = 59;	
	// <<		
  }	  
  
  p->xstring->set_content(p->xstring, newstr);
  for(i=0; i<=j; i++) {
	p->xstring->keycode[i] = keycode[i]; 
	p->xstring->keycodeModifiers[i] = keycodeModifiers[i];		  
  }
  xsm->send_string(p->xstring, p->xswitchlng, p->xevnt, -1);
  
  //xevnt->uninit(xevnt);    
  //xsm->uninit(xsm);
  //free(keycode);
  //free(keycodeModifiers);
  //free(string);
}

/*allocate structs and connect to X*/
xprograminfo *xprogram_init (struct _xprograminfo *toInit, xconf* xnc)
{
  toInit = NULL;
  int i, numGroups;	
  char** names;
  int* nums;
	
  /*allocate & init huge xProgramInfo struct*/
  toInit = (struct _xprograminfo*)xnmalloc ((int) sizeof(struct _xprograminfo));
  if(!toInit) {
    send_error (ERROR, "Bad alloc in xProgramInfo\n", NULL);
    return NULL;
  }
  
  toInit->xobject = xobj_init (toInit->xobject);	/*program window*/
  
  /*connect to X Server*/
  if (ERROR == toInit->xobject->connect (toInit->xobject)) {
	send_error (ERROR, "Connect to XServer Failed\n", NULL);  
    return NULL;
  }
  
  /*set func to catch xlib Errors*/
  toInit->xobject->set_err_handler(handleXErrors);

  toInit->xword = xstr_init(toInit->xword, MAX_WORD_LENGTH);	/*current word flush buffer*/
  toInit->xstring = xstr_init(toInit->xstring, MAX_STRING_LENGTH);	/*current string buffer*/
  toInit->selection = xselection_init(toInit->selection, 
                       toInit->xobject->xDisplay, toInit->xobject->xWindow);	/*X selection processor*/
  toInit->xswitchlng = xswitchlang_init(toInit->xswitchlng, 
                       toInit->xobject->xDisplay);	/*KBD-based layout switch*/
  toInit->xevnt = xevent_init(toInit->xevnt);	/*X event processor*/
  toInit->ifocus = xifocus_init(toInit->ifocus);	/*X Input Focus and Pointer processor*/
  toInit->xklbuffer = xkeylogbuffer_init(toInit->xklbuffer);	/*Input word buffer*/
  toInit->xbtable = xbindingtable_init(toInit->xbtable, 
                   DEFAULT_BIND_TABLE_SIZE);	/*Key Bind Table*/
  
  if (!(toInit->xobject) || !(toInit->xstring) || !(toInit->selection)
      || !(toInit->xevnt) || !(toInit->ifocus)
      || !(toInit->xswitchlng) || !(toInit->xklbuffer)
      || !(toInit->xbtable) || !(toInit->xword)) {
    toInit->xobject->uninit(toInit->xobject);
    toInit->xstring->uninit(toInit->xstring);
    toInit->selection->uninit(toInit->selection);
    toInit->xevnt->uninit(toInit->xevnt);
    toInit->ifocus->uninit(toInit->ifocus);
    toInit->xswitchlng->uninit(toInit->xswitchlng);
    toInit->xklbuffer->free(toInit->xklbuffer);
    toInit->xbtable->uninit(toInit->xbtable);
    
    free (toInit);
    
  
    send_error (ERROR, "Bad alloc in xProgramInfo\n", NULL);
    return NULL;
  }
  
  /*create program window*/
  if (ERROR == toInit->xobject->create_win (toInit->xobject))
    return NULL; 
  /*selecting no core event mask. Remember: selection notify event have no mask.*/
  toInit->xobject->set_event_mask (toInit->xobject, None);
  
  /*xstring & xword xobject init*/
  toInit->xstring->home->xDisplay = toInit->xobject->get_display (toInit->xobject);
  toInit->xstring->home->xWindow = toInit->xobject->get_win(toInit->xobject);
  toInit->xword->home->xDisplay = toInit->xobject->get_display(toInit->xobject);
  toInit->xword->home->xWindow = toInit->xobject->get_win (toInit->xobject);
  /*xselection*/
  toInit->selection->requestor->xDisplay = toInit->xobject->xDisplay;
  toInit->selection->requestor->xWindow = toInit->xobject->get_win (toInit->xobject);
  toInit->selection->owner->xDisplay = toInit->xobject->xDisplay;
  /*xevent*/
  toInit->xevnt->set_communicator (toInit->xevnt, toInit->xobject);
  /*xifocus*/
  toInit->ifocus->owner->xDisplay = toInit->xobject->xDisplay;
  /*xswitching*/
  toInit->xswitchlng->xobject->xDisplay = toInit->xobject->get_display (toInit->xobject);
  toInit->xswitchlng->usa_eng_keyboard_group = xnc->xkbGroup[0];  
  toInit->xswitchlng->rus_win_keyboard_group = xnc->xkbGroup[1];
  names = toInit->xswitchlng->getKeyboardList (&numGroups, toInit->xswitchlng);

  toInit->xswitchlng->currlang = toInit->xswitchlng->get_curr_lang(toInit->xswitchlng, toInit->xobject->xDisplay);
 
  /*xbinding*/
  for(i = 1; i<DEFAULT_BIND_TABLE_SIZE; i++)
    xKeyLogManager_bind (toInit->xbtable, xnc->btable[i].key,
			 xnc->btable[i].key_modifier, i);
  
  XSynchronize (toInit->xobject->xDisplay, TRUE);
  XFlush (toInit->xobject->xDisplay);
  
  // Function mapping
  toInit->uninit = xprogram_uninit; 
  toInit->update = xprogram_update;
  toInit->process_input = xprogram_process_input;
  toInit->perform_current_bind = xprogram_perform_current_bind;
  toInit->perform_current_mode = xprogram_perform_current_mode;
  toInit->chng_word_by_selection = xprogram_chng_word_by_selection;
  toInit->chng_string_by_selection = xprogram_chng_string_by_selection;
  toInit->set_convert_text = xprogram_set_convert_text;
  toInit->get_curr_lang = xprogram_get_curr_lang;
  
  return toInit;
}
