/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Originally written by Art Vodolazsky <burivuh@crew.org.ru>
 *  a lot of modifications and other stuff by 
 *  Alphex Kaänoken <akaanoken@gmail.com>
 *  after Andrew Crew Kuznetsov <andrew.crew.kuznetsov@gmail.com> rewrite it.
 *
 *  (c) Crew IT Research Labs 2004, 2005, 2006
 */

#include "select_manager.h"
#include "../libmisc/error.h" /*FIXME: need to do -I flag on gcc*/
#include "../libmisc/utils.h"

#include <X11/keysym.h>
#include <X11/Xmu/Atoms.h>

#include <unistd.h>
#include <stdlib.h>
#include <string.h>

int xsm_get_string(xselection *sel, xstr *str);
int xsm_get_string(xselection *sel, xstr *str)
{
  char *string = NULL;
  status err = SUCCESS;
  
  string = (char *)sel->get (sel);
  if(!string)
    return ERROR;
  str->set_content (str, NULL);
  err = str->set_content(str, string);
  if(err == BAD_ALLOC) {
    send_error (ERROR, "BAD_ALLOC!\n", NULL);
    return err;
  }  
  free(string);
  
  return err;
}

int xsm_send_string(xstr *str,xswitchlang *swlng, xevent *xevnt,int sw);
int xsm_send_string(xstr *str,xswitchlang *swlng, xevent *xevnt,int sw)
{
  XEvent release_event_backup[MAX_NATIVE_INTERRUPT_EVENTS];
  XEvent press_event_backup[MAX_NATIVE_INTERRUPT_EVENTS];
  status err = SUCCESS;
  num num_of_native_events, index;
  err= swlng->switch_lang(swlng);
  switch (sw) {
    case RUSSIAN:
    case ENGLISH:
      err = str->set_key_code (str, sw, swlng);
      break;
    default:
      send_error (WARNING, "Unknown language to switch!\n", NULL);
      //return WARNING;
      break;
  }
  if (err != SUCCESS) {
    send_error (ERROR, "Can't code string\n", NULL);
    return ERROR;
  }
  num_of_native_events = 0;
  while (xevnt->chck_key_press(xevnt)) {
    if (xevnt->is_native(xevnt)) {
      if (num_of_native_events == MAX_NATIVE_INTERRUPT_EVENTS) {
	    send_error (ERROR,
		    "Too slow process. Have over %d interrupt events!\n",
		    &num_of_native_events);
	    return ERROR;
      }
      release_event_backup[num_of_native_events++] = xevnt->event;
    }
  }
  
  xevnt->send_string (xevnt, str);
  for (index = 0; index < num_of_native_events; index++) {
    xevnt->send_event (xevnt, &press_event_backup[index], CLIENT_EVENT_MASK);
    xevnt->put_back_to_queue (xevnt, &press_event_backup[index]);
  }
  
  return err;
}

int xsm_del_nchrbck(xevent *xevnt,int n);
int xsm_del_nchrbck(xevent *xevnt,int n)
{
  status err = SUCCESS;
  num charIndex;
  
  for(charIndex = 0; charIndex < n; charIndex++){
    err = xevnt->send_xkey(xevnt, XKeysymToKeycode(xevnt->communicator->xDisplay, XK_BackSpace), None);	
    if(err == ERROR){
      send_error(WARNING, "Error sending keys\n", NULL);
      err = WARNING;
      break;
    }
  }
  XFlush (xevnt->communicator->get_display(xevnt->communicator));
  return err;
}

int xsm_sel_allstr_str(xevent *xevnt);
int xsm_sel_allstr_str(xevent *xevnt)
{
  status err = SUCCESS;
  err = xevnt->send_xkey (xevnt, xevnt->keysym_to_char(XK_Home), ShiftMask);
  
  return err;
}

int xsm_sel_allstr_end(xevent *xevnt);
int xsm_sel_allstr_end(xevent *xevnt)
{
  status err = SUCCESS;
  err = xevnt->send_xkey (xevnt, xevnt->keysym_to_char(XK_End), ShiftMask);

  return err;
}

int xsm_sel_nstr_str(xevent *xevnt,int n);
int xsm_sel_nstr_str(xevent *xevnt,int n)
{
  num charIndex;
  status err = SUCCESS;
  
  for (charIndex = 0; charIndex < n; charIndex++) {
    err = xevnt->send_xkey (xevnt, xevnt->keysym_to_char(XK_Left), ShiftMask);
  }
  
  return err;
}

int xsm_sel_nstr_end(xevent *xevnt,int n);
int xsm_sel_nstr_end(xevent *xevnt,int n)
{
  num charIndex;
  status err = SUCCESS;
  
  for (charIndex = 0; charIndex < n; charIndex++) {
    err = xevnt->send_xkey (xevnt, xevnt->keysym_to_char(XK_Right), ShiftMask);
  }
  
  return err;
}

int xsm_usel_nstr_str(xevent *xevnt,int n);
int xsm_usel_nstr_str(xevent *xevnt,int n)
{
  num charIndex;
  status err = SUCCESS;
  
  for (charIndex = 0; charIndex < n; charIndex++) {
    err = xevnt->send_xkey (xevnt, xevnt->keysym_to_char(XK_Left), ShiftMask);
  }
  return err;
}

int xsm_usel_nstr_end(xevent *xevnt,int n);
int xsm_usel_nstr_end(xevent *xevnt,int n)
{
  num charIndex;
  status err = SUCCESS;
  
  for (charIndex = 0; charIndex < n; charIndex++) {
    err = xevnt->send_xkey (xevnt, xevnt->keysym_to_char(XK_Right), ShiftMask);
  }
  
  return err;
}

int xsm_usel_allstr_end(xevent *xevnt);
int xsm_usel_allstr_end(xevent *xevnt)
{
  status err = SUCCESS;
  err = xevnt->send_xkey (xevnt, xevnt->keysym_to_char(XK_End), ShiftMask);
  
  return err;
}

int xsm_uninit(xselectmanager *p);
int xsm_uninit(xselectmanager *p)
{
  status err = SUCCESS;
  //p->str->uninit(p->str);
  free(p);
  return err;
}

/*initialization function*/ /*TODO: rewrite, waiting for library.h*/
xselectmanager* xselectmanager_init(struct _xselectmanager *p)
{
  /*checking for given pointer*/
  if(p!=NULL) {
    fprintf(stderr,"init_xsm: Given pointer is already in use.\n");
    return p;
  }
   
  p = (xselectmanager *)xnmalloc(sizeof(struct _xselectmanager));
  /*init*/
  p->state=0; /*manager state*/
  /*pointers*/

  p->str=xstr_init(p->str, 0);

  p->get_string = xsm_get_string;
  p->send_string = xsm_send_string;
  p->del_nchrbck = xsm_del_nchrbck;
  p->sel_allstr_str = xsm_sel_allstr_str;
  p->sel_allstr_end = xsm_sel_allstr_end;
  p->sel_nstr_str = xsm_sel_nstr_str;
  p->sel_nstr_end = xsm_sel_nstr_end;
  p->usel_allstr_end = xsm_usel_allstr_end;
  p->usel_nstr_str = xsm_usel_nstr_str;
  p->usel_nstr_end = xsm_usel_nstr_end;
  p->uninit = xsm_uninit;

  return p;
}
