/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Originally written by Alphex Kaanoken <kaanoken@crew.org.ru>
 *  and Art Vodolazsky <burivuh@crew.org.ru>
 *
 *  NOTE: This is init file of X Neural Switcher
 *
 *  (c) Crew IT Research Labs 2004
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>

#include "libmisc/xconfig.h"
#include "libx/program.h"
#include "libmisc/utils.h"

#define CNF_DIR   ".xneur"
#define RUDICT_NAME "rudict"
#define ENDICT_NAME "endict"
#define CNF_NAME  "xneurrc"
#define MIN_LEN_TO_WORK 1
#define LOCK_FILE   "/tmp/xneur.pid"

int get_xneur_pid (char* arg[])
{
	DIR* proc = opendir("/proc");
	if (proc == NULL)
	{
		printf("failed to open /proc\n");
		return -1;
	}
	struct dirent* ent;
	int count_xneur_proc = 0;
	int xneur_pid = 0;
	while (ent = readdir(proc))
	{
		if (!isdigit(ent->d_name[0]))
			continue;
		char buffer[NAME_MAX] = "/proc/";
		strcat(buffer, ent->d_name);
		strcat(buffer, "/cmdline");
		FILE* cmd = fopen(buffer, "r");
		if (cmd == NULL)
		{
			printf("failed to open file '%s'\n", buffer);
			continue;
		}
		int rd = fread(buffer, 1, NAME_MAX, cmd);
		buffer[rd] = '\0';
		fclose(cmd);
		if ((strstr(buffer, "xneur") == NULL) || (strstr(buffer, "xneur") != buffer))
			continue;
		// xneur PID found
		count_xneur_proc++;
		xneur_pid = atoi(ent->d_name);
	}
	closedir(proc);
	if (count_xneur_proc == 0) {
	  send_error(ERROR, "xneur PID is not found\n", NULL);
	  return 0;
	}
	else if (count_xneur_proc == 1) {
	  printf("found xneur PID: %d\n", xneur_pid);
	  return xneur_pid;
	}
	else {
	  printf("found other xneur PID!\n");
	  return -1;
	}
}

int set_lock (char* arg[])
{
	int xneur_pid = 0;
	FILE* fpid = fopen(LOCK_FILE, "w");
	if (!fpid)
	{
		fprintf(stderr, "%s: failed to create lock file.\n", arg[0]);
		return 0;
	}	
	xneur_pid = get_xneur_pid(arg);
	printf("Set PID to lock file...\n");
	if (xneur_pid > 0) {
	  fprintf(fpid, "%d", xneur_pid);
	  fclose(fpid);
	  return 1;	
	}
	else {
	  fprintf(fpid, "%d", 0);
	  fclose(fpid);
	  return 0;	
	}
}

int main (int argc, char *argv[])
{  
  bool have_client = TRUE, to_continue;
  app_list *supported_app_list = NULL;
  size word_len = 0;
  const char* home = getenv("HOME");
  int cnf_len=strlen(home)+strlen(CNF_DIR)+strlen(CNF_NAME)+3;  
  char *config_file=(char *) xnmalloc(cnf_len);
  char *rudict_file;
  char *endict_file;
  xprograminfo *xinfo=NULL;
  xconf *xnc = NULL;

  // Set locking
  if (!set_lock(argv))
	exit(0);
	
  config_file=memset(config_file,'\0',cnf_len);  
  snprintf (config_file,cnf_len,"%s/%s/%s",getenv ("HOME"),CNF_DIR,CNF_NAME);
  // Dictionary Files 	
  rudict_file = malloc (cnf_len + strlen (RUDICT_NAME) + 1);
  endict_file = malloc (cnf_len + strlen (ENDICT_NAME) + 1);
  sprintf (rudict_file, "%s/%s/%s", home, CNF_DIR, RUDICT_NAME);
  sprintf (endict_file, "%s/%s/%s", home, CNF_DIR, ENDICT_NAME); 
  // load config structure 
  xnc=xconf_init(xnc);
  send_error(DEBUG,"Initial Configuration Complete!\n",NULL);
  xnc->load_file(xnc,config_file);
  send_error(DEBUG,"Loading Configuration File Complete!\n",NULL);	
  free(config_file);
  // load dictionary files
  xnc->rudict = xnc->load_dict_file (rudict_file);
  send_error (DEBUG, "Russian Dictionary complete\n", NULL);
  xnc->endict = xnc->load_dict_file (endict_file);	
  send_error (DEBUG, "English Dictionary complete\n", NULL);

  //init program structures
  xinfo=xprogram_init(xinfo, xnc);
  send_error(DEBUG,"Init Program Structure Complete!\n",NULL);
  if(xinfo==NULL)
    xprogram_exit();
  // init current mode value
  xnc->CurrentMode=xnc->DefaultMode;
  send_error(DEBUG,"Current mode is set to %d\n",&(xnc->CurrentMode));
  //send_error(DEBUG,"Current language detect mode is set to %d\n", &(xnc->LangDetectMode));
  
  //just for testing
  if(argv[1]){
	  printf("Loaded string '%s' len=%d.\n", (const char*)argv[1], strlen((const char*)argv[1]));
	  char * string = xnmalloc(strlen((const char*)argv[1]));
	  strcpy(string, (const char*)argv[1]);
	  int i = 0, j = -1;
	  char * cyrstr = xnmalloc(3);
	  char * newstr = xnmalloc(MAX_STRING_LENGTH);
	  newstr = memset (newstr, NULLSYM, MAX_WORD_LENGTH);
	  for(i=0; i<strlen(string); i++) {
		if (isascii(string[i])) {
		  printf("Is ASCII\n");
		  newstr[++j]=string[i];
		}	
		else {
		  printf("Is CYR. ");	
		  cyrstr[0]=string[i];
		  i++;
		  cyrstr[1]=string[i];
		  cyrstr[2]='\0';
		  printf("ASCII '%c'\n", get_ascii(cyrstr));
		  newstr[++j]=get_ascii(cyrstr);
		}	
	  }	  
	  printf("New string '%s' len=%d.\n", newstr, strlen(newstr));
	  printf("Exit.\n");
    return 0;
  }


  while (!xprogram_isexiting()) {
    // pause between iterations
    usleep(xnc->ScaningDelay);
    supported_app_list=xnc->alist;
    // update current xclient info
    have_client=xinfo->update(xinfo,xnc->alist);
    if (have_client==FALSE)
      continue;
    // so client has met our conditions. Let's parse input keys
    word_len=xinfo->process_input(xinfo);
    // perform bind actions if any			
	to_continue=xinfo->perform_current_bind (xinfo, xnc);			 
    if(to_continue)
      continue;
    // so we have no bind during this iteration. Perform mode actions
    // if any.
	
    if(word_len>=MIN_LEN_TO_WORK)
      xinfo->perform_current_mode(xinfo,xnc);
  }
  
  //xnc->uninit(xnc);
  send_error(DEBUG,"Current config is free!\n",NULL);
  //xinfo->uninit(xinfo);
  send_error(DEBUG,"Current program info is free!\n",NULL);
  return 0;
}
